@property
def L(self) -> float:
    """liter"""
    return 1e-3 * self.m**3

@property
def mL(self) -> float:
    """milliliter"""
    return 1e-3 * self.L

@property
def uL(self) -> float:
    """microliter"""
    return 1e-6 * self.L

@property
def nL(self) -> float:
    """nanoliter"""
    return 1e-9 * self.L

@property
def pL(self) -> float:
    """picoliter"""
    return 1e-12 * self.L

@property
def fL(self) -> float:
    """femtoliter"""
    return 1e-15 * self.L

@property
def aL(self) -> float:
    """attoliter"""
    return 1e-18 * self.L

@property
def kL(self) -> float:
    """kiloliter"""
    return 1e3 * self.L

@property
def ML(self) -> float:
    """megaliter"""
    return 1e6 * self.L

@property
def GL(self) -> float:
    """gigaliter"""
    return 1e9 * self.L


# coding: utf-8
# Standard Python libraries
from pathlib import Path
import shutil
import tarfile
from collections import OrderedDict
from typing import Optional, Tuple, Union

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://api.mongodb.com/python/current/
import pymongo
from pymongo import MongoClient
from gridfs import GridFS

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# Relative imports
from ..tools import aslist
from . import Database
from ..record import recordmanager, load_record, Record

class MongoDatabase(Database):

    def __init__(self,
                 host: str = 'localhost',
                 port: int = 27017,
                 database: str = 'iprPy',
                 **kwargs):
        """
        Initializes a connection to a Mongo database.
        
        Parameters
        ----------
        host : str, optional
            The mongo host to connect to.  Default value is 'localhost'.
        port : int, optional
            Then port to use in connecting to the mongo host.  Default value
            is 27017.
        database : str, optional
            The name of the database in the mongo host to interact with.
            Default value is 'iprPy'
        **kwargs : dict, optional
            Any extra keyword arguments needed to initialize a
            pymongo.MongoClient object.
        """

        # Connect to underlying class
        self.__mongodb = MongoClient(host=host, port=port, document_class=DM, **kwargs)[database]

        # Define class host using client's host, port and database name
        host = f'{host}:{port}.{database}'

        # Pass host to Database initializer
        Database.__init__(self, host)

    @property
    def style(self) -> str:
        """str: The database style"""
        return 'mongo'

    @property
    def mongodb(self) -> pymongo.database.Database:
        """pymongo.Database : The underlying database API object."""
        return self.__mongodb

    def get_records(self,
                    style: Optional[str] = None,
                    return_df: bool = False,
                    query: Optional[dict] = None,
                    **kwargs) -> Union[list, Tuple[list, pd.DataFrame]]:
        """
        Produces a list of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search.  If not given, a prompt will ask for it.
        return_df : bool, optional
            If True, then the corresponding pandas.Dataframe of metadata
            will also be returned
        query : dict, optional
            A custom-built Mongo-style query to use for the record search.
            Alternative to passing in the record-specific metadata kwargs.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        ------
        records : numpy.NDArray
            All records from the database matching the given parameters.
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.  Only returned
            if return_df is True.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Use given query
        if query is not None:
            assert len(kwargs) == 0, 'query cannot be given with kwargs'
        else:
            query = load_record(style).mongoquery(**kwargs)

        # Query the collection to construct records
        records = []
        collection = self.mongodb[style]
        for entry in collection.find(query):
            record = load_record(style, model=entry['content'],
                                 name=entry['name'], database=self)
            records.append(record)
        records = np.array(records)

        # Build df
        if len(records) > 0:
            df = []
            for record in records:
                df.append(record.metadata())
            df = pd.DataFrame(df)
        else:
            df = pd.DataFrame({'name':[]})

        # Sort by name
        df = df.sort_values('name')
        records = records[df.index.tolist()]

        # Return records (and df)
        if return_df:
            return records, df.reset_index(drop=True)
        else:
            return records

    def get_records_df(self,
                       style: Optional[str] = None,
                       query: Optional[dict] = None,
                       **kwargs) -> pd.DataFrame:
        """
        Produces a pandas.Dataframe of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search.
        query : dict, optional
            A custom-built Mongo-style query to use for the record search.
            Alternative to passing in the record-specific metadata keywords.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        -------
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.
        """
        return self.get_records(style, query=query, return_df=True, **kwargs)[1]

    def get_record(self,
                   style: Optional[str] = None,
                   query: Optional[dict] = None,
                   **kwargs) -> Record:
        """
        Retrieves a single matching record from the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search.
        query : dict, optional
            A custom-built Mongo-style query to use for the record search.
            Alternative to passing in the record-specific metadata keywords.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        ------
        Record
            The record from the database matching the given parameters.
        
        Raises
        ------
        ValueError
            If multiple or no matching records found.
        """

        if style is None:
            styles = recordmanager.loaded_style_names
        else:
            styles = aslist(style)

        # Get records
        records = []
        for style in styles:
            records.append(self.get_records(style, query=query, **kwargs))
        records = np.hstack(records)

        # Verify that there is only one matching record
        if len(records) == 1:
            return records[0]
        elif len(records) == 0:
            raise ValueError('No matching records found')
        else:
            raise ValueError('Multiple matching records found')

    def count_records(self,
                      style: Optional[str] = None,
                      query: Optional[dict] = None,
                      **kwargs) -> int:
        """
        Retrieves a count of matching records from the database.  Much faster
        than get_records if you only want to know the number of matches.
        
        Parameters
        ----------
        style : str, optional
            The record style to search.
        query : dict, optional
            A custom-built Mongo-style query to use for the record search.
            Alternative to passing in the record-specific metadata keywords.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        ------
        int
            The count of records in the database matching the given parameters.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Use given query
        if query is not None:
            assert len(kwargs) == 0, 'query cannot be given with kwargs'
        else:
            query = load_record(style).mongoquery(**kwargs)

        # Query the collection to construct records
        collection = self.mongodb[style]
        count = collection.count_documents(query)

        return count

    def add_record(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   model: Union[str, DM, None] = None,
                   build: bool = False,
                   verbose: bool = False) -> Record:
        """
        Adds a new record to the database.
        
        Parameters
        ----------
        record : Record, optional
            The new record to add to the database.  If not given, then name,
            style and content are required.
        style : str, optional
            The record style for the new record.  Required if record is not
            given.
        name : str, optional
            The name to assign to the new record.  Required if record is not
            given.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.

        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and content.
        
        Raises
        ------
        ValueError
            If style, name and/or content given with record, or a matching record
            already exists.
        """

        # Create Record object if not given
        if record is None:
            record = load_record(style, model=model, name=name)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None or model is not None:
            raise ValueError('kwargs style, name, and content cannot be given with kwarg record')

        # Verify that there isn't already a record with a matching name
        if len(self.get_records(name=record.name, style=record.style)) > 0:
            raise ValueError(f'Record {record.name} already exists')

        # Retrieve/build model contents
        try:
            assert build is False
            model = record.model
        except:
            model = record.build_model()

        # Create meta mongo entry
        entry = OrderedDict()
        entry['name'] = record.name
        entry['content'] = model

        # Upload to mongodb
        self.mongodb[record.style].insert_one(entry)

        if verbose:
            print(f'{record} added to {self.host}')

        return record

    def update_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      model: Union[str, DM, None] = None,
                      build: bool = False,
                      verbose: bool = False) -> Record:
        """
        Replaces an existing record with a new record of matching name and
        style, but new content.
        
        Parameters
        ----------
        record : Record, optional
            The record with new content to update in the database.  If not
            given, content is required along with name and/or style to
            uniquely define a record to update.
        style : str, optional
            The style of the record to update.
        name : str, optional
            The name to uniquely identify the record to update.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        
        Returns
        ------
        Record
            Either the given record or a record composed of the name, style,
            and content.
        
        Raises
        ------
        TypeError
            If no new content is given.
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            if model is None:
                raise TypeError('no new model given')
            oldrecord = self.get_record(name=name, style=style)
            record = load_record(oldrecord.style, model=model, name=oldrecord.name)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Replace model in record object
        elif model is not None:
            oldrecord = record
            record = load_record(oldrecord.style, model=model, name=oldrecord.name)

        # Find oldrecord matching record
        else:
            oldrecord = self.get_record(name=record.name, style=record.style)

        # Delete oldrecord
        self.delete_record(record=oldrecord)

        # Add new record
        self.add_record(record=record, build=build)

        if verbose:
            print(f'{record} updated in {self.host}')

        return record

    def delete_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      verbose: bool = False):
        """
        Permanently deletes a record from the database. 
        
        Parameters
        ----------
        record : Record, optional
            The record to delete from the database.  If not given, name and/or
            style are needed to uniquely define the record to delete.
        name : str, optional
            The name of the record to delete.
        style : str, optional
            The style of the record to delete.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.

        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Build delete query
        query = {}
        query['name'] = record.name

        # Delete record 
        self.mongodb[record.style].delete_one(query)

        if verbose:
            print(f'{record} deleted from {self.host}')

    def add_tar(self, 
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                tar: Optional[bytes] = None,
                root_dir: Optional[Path] = None):
        """
        Archives and stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        name : str, optional
            .The name to use in uniquely identifying the record.
        style : str, optional
            .The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has an archive.
        """
        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Verify that record exists
        else:
            record = self.get_record(name=record.name, style=record.style)

        # Define mongofs
        mongofs = GridFS(self.mongodb, collection=record.style)

        # Check if an archive already exists
        if mongofs.exists({"recordname": record.name}):
            raise ValueError('Record already has an archive')

        if tar is None:
            if root_dir is None:
                root_dir = Path.cwd()

            # Make archive
            basename = Path(root_dir, record.name)
            filename = Path(root_dir, record.name + '.tar.gz')
            shutil.make_archive(basename, 'gztar', root_dir=root_dir,
                                base_dir=record.name)

            # Upload archive
            with open(filename, 'rb') as f:
                tries = 0
                while tries < 2:
                    tries += 1
                    try:
                        mongofs.put(f, recordname=record.name)
                        break
                    except Exception as err:
                        if tries == 2:
                            raise ValueError('Failed to upload archive 2 times') from err

            # Remove local archive copy
            filename.unlink()

        elif root_dir is None:
            # Upload archive
            tries = 0
            while tries < 2:
                tries += 1
                try:
                    mongofs.put(tar, recordname=record.name)
                    break
                except Exception as err:
                    if tries == 2:
                        raise ValueError('Failed to upload archive 2 times') from err
        else:
            raise ValueError('tar and root_dir cannot both be given')

    def get_tar(self,
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                raw: bool = False) -> Union[tarfile.TarFile, bytes]:
        """
        Retrieves the tar archive associated with a record in the database.
                
        Parameters
        ----------
        record : Record, optional
            The record to retrieve the associated tar archive for.
        name : str, optional
            .The name to use in uniquely identifying the record.
        style : str, optional
            .The style to use in uniquely identifying the record.
        raw : bool, optional
            If True, return the archive as raw binary content. If 
            False, return as an open tarfile. (Default is False)
        
        Returns
        -------
        tarfile or str
            The tar archive as an open tarfile if raw=False, or as a binary str if
            raw=True.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a TypeError for competing kwargs
        elif style is not None or name is not None:
            raise TypeError('kwargs style and name cannot be given with kwarg record')

        # Define mongofs
        mongofs = GridFS(self.mongodb, collection=record.style)

        # Build query
        query = {}
        query['recordname'] = record.name

        # Get tar
        matches = list(mongofs.find(query))
        if len(matches) == 1:
            tar = matches[0]
        elif len(matches) == 0:
            raise ValueError('No tar found for the record')
        else:
            raise ValueError('Multiple tars found for the record')

        # Return content
        if raw is True:
            return tar.read()
        else:
            tarobj = tarfile.open(fileobj=tar)
            record.tar = tarobj
            return tarobj

    def delete_tar(self, 
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None):
        """
        Deletes a tar file from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record associated with the tar archive to delete.  If not
            given, then name and/or style necessary to uniquely identify
            the record are needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None:
            raise ValueError('kwargs style and name cannot be given with kwarg record')

        # Define mongofs
        mongofs = GridFS(self.mongodb, collection=record.style)

        # Build query
        query = {}
        query['recordname'] = record.name

        # Get tar
        matches = list(mongofs.find(query))
        if len(matches) == 1:
            tar = matches[0]
        elif len(matches) == 0:
            raise ValueError('No tar found for the record')
        else:
            raise ValueError('Multiple tars found for the record')

        # Delete tar
        mongofs.delete(tar._id)

    def update_tar(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   tar: Optional[bytes] = None,
                   root_dir: Optional[Path] = None):
        """
        Replaces an existing tar archive for a record with a new one. 
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then 
            name and/or style necessary to uniquely identify the record are 
            needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        """

        # Delete the existing tar archive stored in the database
        self.delete_tar(record=record, name=name, style=style)

        # Add the new tar archive
        self.add_tar(record=record, name=name, style=style, tar=tar, root_dir=root_dir)



@property
def F(self) -> float:
    """farad"""
    return self.C / self.V

@property
def uF(self) -> float:
    """microfarad"""
    return 1e-6 * self.F

@property
def nF(self) -> float:
    """nanofarad"""
    return 1e-9 * self.F

@property
def pF(self) -> float:
    """picofarad"""
    return 1e-12 * self.F

@property
def fF(self) -> float:
    """femtofarad"""
    return 1e-15 * self.F

@property
def aF(self) -> float:
    """attofarad"""
    return 1e-18 * self.F

@property
def H(self) -> float:
    """henry"""
    return self.m**2 * self.kg / self.C**2

@property
def mH(self) -> float:
    """millihenry"""
    return 1e-3 * self.H

@property
def uH(self) -> float:
    """microhenry"""
    return 1e-6 * self.H

@property
def nH(self) -> float:
    """nanohenry"""
    return 1e-9 * self.H


@property 
def T(self) -> float:
    """tesla"""
    return (self.V * self.s) / self.m**2

@property
def mT(self) -> float:
    """millitesla"""
    return 1e-3 * self.T

@property
def uT(self) -> float:
    """microtesla"""
    return 1e-6 * self.T

@property
def nT(self) -> float:
    """nanotesla"""
    return 1e-9 * self.T

@property
def G(self) -> float:
    """gauss"""
    return 1e-4 * self.T

@property
def mG(self) -> float:
    """milligauss"""
    return 1e-3 * self.G

@property
def uG(self) -> float:
    """microgauss"""
    return 1e-6 * self.G

@property
def kG(self) -> float:
    """kilogauss"""
    return 1e3 * self.G

@property
def Oe(self) -> float:
    """oersted"""
    return (1000. / (4. * self.pi)) * self.A / self.m

@property
def Wb(self) -> float:
    """weber"""
    return self.J / self.A


# coding: utf-8

# Standard Python libraries
from typing import Any, Optional

import pandas as pd
class Query():
    """
    Base Query class.  Each Query class defines a query operation and each
    Query object is associated with a specific data field.
    """
    
    def __init__(self,
                 name: Optional[str] = None,
                 parent: Optional[str] = None,
                 path: Optional[str] = None,
                 description: str = ''):
        """
        Query initialization

        Parameters
        ----------
        name : str or None, optional
            The metadata key associated with the data field.  Must be set
            to use the pandas query method.
        parent : str or None, optional
            Allows for the pandas query operations to work on embedded
            metadata dicts.  If given, the pandas query method will check the
            value of metadata[parent][name].
        path : str or None, optional
            The record data path to the data field.  Levels are delimited by
            periods.  Must be given to use the mongo query method.
        description : str, optional
            Description of the query operation, i.e. what it is searching.
        """
        # Check that object is a subclass
        if self.__module__ == __name__:
            raise TypeError("Don't use Query itself, only use derived classes")

        self.name = name
        self.parent = parent
        self.path = path
        self.description = description

    def __str__(self) -> str:
        """
        Returns
        -------
        str
            The string representation of the query.
        """
        return f'query style {self.style}'

    @property
    def style(self) -> str:
        """str: The query style"""
        raise NotImplementedError('Not defined for base class')

    @property
    def parameter_type(self) -> str:
        """str: The types of query parameter values accepted by this query style"""
        return 'str or list, optional'

    @property
    def name(self) -> str:
        """str: The metadata key associated with the data field"""
        if self.__name is None:
            raise AttributeError('name not set')
        return self.__name

    @name.setter
    def name(self, value: Optional[str]):
        if value is None:
            self.__name = None
        elif isinstance(value, str):
            self.__name = value
        else:
            raise TypeError('name must be None or a string')

    @property
    def parent(self) -> Optional[str]:
        """str or None: The parent metadata key, if any."""
        return self.__parent

    @parent.setter
    def parent(self, value: Optional[str]):
        if value is None:
            self.__parent = None
        elif isinstance(value, str):
            self.__parent = value
        else:
            raise TypeError('parent must be None or a string')

    @property
    def path(self) -> str:
        """str: The period delimited path to the associated field."""
        if self.__path is None:
            raise AttributeError('path not set')
        return self.__path

    @path.setter
    def path(self, value: Optional[str]):
        if value is None:
            self.__path = None
        elif isinstance(value, str):
            self.__path = value
        else:
            raise TypeError('path must be None or a string')

    @property
    def description(self) -> str:
        """str: Describes the query operation"""
        return self.__description
    
    @description.setter
    def description(self, value: str):
        self.__description = str(value)

    def mongo(self,
              querylist: list,
              value: Any,
              prefix: str = ''):
        """
        Builds a Mongo query operation for the field.

        Parameters
        ----------
        querylist : list
            The working list of mongo query operations which is to be appended
            with the operation for this query object.
        value : any
            The value of the field to query on.  If None, then no new query
            operation will be added.
        prefix : str, optional
            An optional prefix to add before the query path.  Used by Record's
            mongoquery to start each path with "content."
        """
        # Do nothing - base class
        pass

    def pandas(self,
               df: pd.DataFrame, 
               value: Any) -> pd.Series:
        """
        Applies a query filter to the metadata for the field.
        
        Parameters
        ----------
        df : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        value : any
            The value of the field to query on.  If None, then it should return
            True for all rows of df.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        # Do nothing - base class
        return df.apply(lambda series:True, axis=1)



# coding: utf-8
# Standard Python libraries
from pathlib import Path
import shutil
import tarfile
from io import BytesIO
from typing import Optional, Tuple, Union

# https://github.com/usnistgov/pycdcs
from cdcs import CDCS

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# Relative imports
from ..tools import aslist, iaslist
from . import Database
from ..record import recordmanager, load_record, Record

class CDCSDatabase(Database):

    def __init__(self,
                 host: str,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 auth: Optional[Tuple[str]] = None,
                 cert: Union[str, Tuple[str], None] = None, 
                 certification: Union[str, Tuple[str], None] = None,
                 verify: Optional[bool] = True,
                 cdcsversion: Optional[str] = None):
        """
        Initializes a database of style curator.
        
        Parameters
        ----------
        host : str
            The host name (url) for the database.
        username : str, optional
            Username of desired account on the server. A prompt will ask for
            the username if not given.
        password : str, optional
            Password of desired account on the server.  This can either be the
            password as a str, or a str path to a file containing only the
            password.  A prompt will ask for the password if not given.
        auth : tuple, optional
            Auth tuple to enable Basic/Digest/Custom HTTP Auth.  Alternative to
            giving username and password separately.
        cert : str, optional
            if String, path to ssl client cert file (.pem). If Tuple,
            ('cert', 'key') pair.
        certification : str, optional
            Alias for cert. Retained for compatibility.
        verify : bool or str, optional
            Either a boolean, in which case it controls whether we verify the
            server's TLS certificate, or a string, in which case it must be a
            path to a CA bundle to use. Defaults to True.
        cdcsversion : str, optional
            For CDCS versions 2.X.X, this allows for specifying the full CDCS
            version to ensure the class methods perform the correct REST
            calls.  This can be specified as "#.#.#", or if None is given will
            default to "2.15.0".  For CDCS versions 3.X.X, this is ignored as
            version info is obtained directly from the database.
        """
        # Fetch password from file if needed
        try:
            with open(password, encoding='UTF-8') as f:
                password = f.read().strip()
        except Exception:
            pass

        # Pass parameters to cdcs object
        self.__cdcs = CDCS(host, username=username, password=password, auth=auth,
                           cert=cert, certification=certification, verify=verify,
                           cdcsversion=cdcsversion)

        # Pass host to Database initializer
        Database.__init__(self, host)

    @property
    def style(self) -> str:
        """str: The database style"""
        return 'cdcs'

    @property
    def cdcs(self) -> CDCS:
        """cdcs.CDCS : The underlying database API object."""
        return self.__cdcs

    def get_records(self,
                    style: Optional[str] = None,
                    return_df: bool = False,
                    name: Union[str, list, None] = None,
                    query: Optional[dict] = None,
                    keyword: Optional[str] = None,
                    **kwargs) -> Union[list, Tuple[list, pd.DataFrame]]:
        """
        Produces a list of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search. If not given, a prompt will ask for it.
        name : str or list, optional
            Record name(s) to delimit by. 
        return_df : bool, optional
            If True, then the corresponding pandas.Dataframe of metadata
            will also be returned
        query : dict, optional
            A custom-built CDCS-style query to use for the record search.
            Alternative to passing in the record-specific metadata kwargs.
            Note that name can be given with query.
        keyword : str, optional
            Allows for a search of records whose contents contain a keyword.
            Alternative to giving query or kwargs.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
        
        Returns
        ------
        records : numpy.NDArray
            All records from the database matching the given parameters.
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.  Only returned
            if return_df is True.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Setup keyword search
        if keyword is not None:
            assert len(kwargs) == 0, 'keyword cannot be given with kwargs'
            assert query is None, 'keyword cannot be given with query'

        # Setup query
        elif query is not None:
            assert len(kwargs) == 0, 'query cannot be given with kwargs'
        else:
            query = load_record(style).cdcsquery(**kwargs)

        def build_records(series):
            return load_record(series.template_title, model=series.xml_content,
                               name=series.title, database=self)

        # Build records by querying for each record name (or None)
        records = []
        for n in iaslist(name):
            data = self.cdcs.query(title=n, template=style, mongoquery=query, keyword=keyword)
            if len(data) > 0:
                records.extend(data.apply(build_records, axis=1))
        records = np.array(records)

        # Build df
        if len(records) > 0:
            df = []
            for record in records:
                df.append(record.metadata())
            df = pd.DataFrame(df)
        else:
            df = pd.DataFrame({'name':[]})

        # Sort by name
        df = df.sort_values('name')
        records = records[df.index.tolist()]

        # Return records (and df)
        if return_df:
            return records, df.reset_index(drop=True)
        else:
            return records

    def get_records_df(self,
                       style: Optional[str] = None,
                       name: Union[str, list, None] = None,
                       query: Optional[dict] = None,
                       keyword: Optional[str] = None,
                       **kwargs) -> pd.DataFrame:
        """
        Produces a list of all matching records in the database.
        
        Parameters
        ----------
        style : str, optional
            The record style to search. If not given, a prompt will ask for it.
        name : str or list, optional
            Record name(s) to delimit by. 
        query : dict, optional
            A custom-built CDCS-style query to use for the record search.
            Alternative to passing in the record-specific metadata kwargs.
            Note that name can be given with query.
        keyword : str, optional
            Allows for a search of records whose contents contain a keyword.
            Alternative to giving query or kwargs.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        -------
        records_df : pandas.DataFrame
            The corresponding metadata values for the records.
        """
        return self.get_records(style, name=name, query=query, keyword=keyword, return_df=True, **kwargs)[1]

    def get_record(self, 
                   style: Optional[str] = None,
                   name: Union[str, list, None] = None,
                   query: Optional[dict] = None,
                   keyword: Optional[str] = None,
                   **kwargs) -> Record:
        """
        Returns a single matching record from the database.
        
        Parameters
        ----------
        style : str, optional
            Record style(s) to limit the search by.
        name : str or list, optional
            Record name(s) to delimit by. 
        query : dict, optional
            A custom-built CDCS-style query to use for the record search.
            Alternative to passing in the record-specific metadata kwargs.
            Note that name can be given with query.
        keyword : str, optional
            Allows for a search of records whose contents contain a keyword.
            Alternative to giving query or kwargs.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
            
        Returns
        ------
        Record
            The single record from the database matching the given parameters.
        
        Raises
        ------
        ValueError
            If multiple or no matching records found.
        """
        if style is None:
            styles = recordmanager.loaded_style_names
        else:
            styles = aslist(style)

        # Get records
        records = []
        for style in styles:
            records.append(self.get_records(style, name=name, query=query, keyword=keyword, **kwargs))
        records = np.hstack(records)

        # Verify that there is only one matching record
        if len(records) == 1:
            return records[0]
        elif len(records) == 0:
            raise ValueError('No matching records found')
        else:
            raise ValueError('Multiple matching records found')
    
    def count_records(self,
                      style: Optional[str] = None,
                      name: Union[str, list, None] = None,
                      query: Optional[dict] = None,
                      keyword: Optional[str] = None,
                      **kwargs) -> int:
        """
        Retrieves a count of matching records from the database.  Much faster
        than get_records if you only want to know the number of matches.
        
        Parameters
        ----------
        style : str, optional
            The record style to search. If not given, a prompt will ask for it.
        name : str or list, optional
            Record name(s) to delimit by. 
        query : dict, optional
            A custom-built CDCS-style query to use for the record search.
            Alternative to passing in the record-specific metadata kwargs.
            Note that name can be given with query.
        keyword : str, optional
            Allows for a search of records whose contents contain a keyword.
            Alternative to giving query or kwargs.
        **kwargs : any, optional
            Any of the record-specific metadata keywords that can be searched
            for.
        
        Returns
        ------
        int
            The count of records in the database matching the given parameters.
        """
        # Set default search parameters
        if style is None:
            style = self.select_record_style()

        # Setup keyword search
        if keyword is not None:
            assert len(kwargs) == 0, 'keyword cannot be given with kwargs'
            assert query is None, 'keyword cannot be given with query'

        # Setup query
        elif query is not None:
            assert len(kwargs) == 0, 'query cannot be given with kwargs'
        else:
            query = load_record(style).cdcsquery(**kwargs)

        count = 0
        for n in iaslist(name):
            count += self.cdcs.query_count(title=n, template=style, mongoquery=query, keyword=keyword)

        return count

    def add_record(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   model: Union[str, DM, None] = None,
                   build: bool = False,
                   verbose: bool = False,
                   workspace: Union[str, pd.Series, None] = None,
                   auto_set_pid_off: bool = False) -> Record:
        """
        Adds a new record to the database.
        
        Parameters
        ----------
        record : iprPy.Record, optional
            The new record to add to the database.  If not given, then name,
            style and content are required.
        style : str, optional
            The record style for the new record.  Required if record is not
            given.
        name : str, optional
            The name to assign to the new record.  Required if record is not
            given.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        workspace : str or pandas.Series, optional
            The name of a workspace to assign the record to.  If not given
            then the record is not assigned to a workspace and will only be
            accessible to the user who uploaded it.
        auto_set_pid_off : bool
            If True, the database's auto_set_pid setting will temporarily be
            turned off for the record upload.  This is only needed if the
            record contains a PID URL corresponding to a set PID xpath.  If
            uploading multiple records, it is better to use the CDCS object's
            auto_set_pid_off() context manager or to manually change the
            setting off/on with CDCS.auto_set_pid.

        Returns
        ------
        Record
            Either the given record or a record composed of the name, style, 
            and model.
        
        Raises
        ------
        ValueError
            If style, name and/or model given with record, or a matching record
            already exists.
        """

        # Create Record object if not given
        if record is None:
            record = load_record(style, model=model, name=name)

        # Issue a ValueError for competing kwargs
        elif style is not None or name is not None or model is not None:
            raise ValueError('kwargs style, name, and model cannot be given with kwarg record')

        # Retrieve/build model contents
        try:
            assert build is False
            content = record.model.xml()
        except Exception:
            content = record.build_model().xml()

        # Upload to database
        self.cdcs.upload_record(template=record.style, content=content,
                                title=record.name,
                                auto_set_pid_off=auto_set_pid_off)
        if verbose:
            print(f'{record} added to {self.host}')

        if workspace is not None:
            self.assign_records(record, workspace, verbose=verbose)

        return record

    def update_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      model: Union[str, DM, None] = None,
                      build: bool = False,
                      verbose: bool = False,
                      workspace: Union[str, pd.Series, None] = None,
                      auto_set_pid_off: bool = False) -> Record:
        """
        Replaces an existing record with a new record of matching name and 
        style, but new content.
        
        Parameters
        ----------
        record : Record, optional
            The record with new content to update in the database.  If not
            given, content is required along with name and/or style to
            uniquely define a record to update.
        style : str, optional
            The style of the record to update.
        name : str, optional
            The name to uniquely identify the record to update.
        model : str or DataModelDict, optional
            The model contents of the new record.  Required if record is not
            given.
        build : bool, optional
            If True, then the uploaded content will be (re)built based on the
            record's attributes.  If False (default), then record's existing
            content will be loaded if it exists, or built if it doesn't exist.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        workspace : str or pandas.Series, optional
            The name of a workspace to assign the record to.  If not given
            then the record is not assigned to a workspace and will only be
            accessible to the user who uploaded it.
        auto_set_pid_off : bool
            If True, the database's auto_set_pid setting will temporarily be
            turned off for the record upload.  This is only needed if the
            record contains a PID URL corresponding to a set PID xpath.  If
            uploading multiple records, it is better to use the CDCS object's
            auto_set_pid_off() context manager or to manually change the
            setting off/on with CDCS.auto_set_pid.
        Returns
        ------
        Record
            Either the given record or a record composed of the name, style, 
            and model.
            
        Raises
        ------
        TypeError
            If no new content is given.
        ValueError
            If style, model, and/or name given with record.
        """

        # Create Record object if not given
        if record is None:
            if model is None:
                raise TypeError('no new model given')
            oldrecord = self.get_record(name=name, style=style)
            record = load_record(oldrecord.style, model=model, name=oldrecord.name)

        # Use given record object
        else:
            if style is not None or name is not None:
                raise ValueError('kwargs style and name cannot be given with kwarg record')

            # Replace model in record object
            if model is not None:
                record = load_record(record.style, model=model, name=record.name)

        # Retrieve/build model contents
        try:
            assert build is False
            content = record.model.xml()
        except Exception:
            content = record.build_model().xml()

        # Upload to database
        self.cdcs.update_record(template=record.style, content=content,
                                title=record.name,
                                auto_set_pid_off=auto_set_pid_off)

        if verbose:
            print(f'{record} updated in {self.host}')

        if workspace is not None:
            self.assign_records(record, workspace, verbose=verbose)

        return record

    def delete_record(self,
                      record: Optional[Record] = None,
                      style: Optional[str] = None,
                      name: Optional[str] = None,
                      verbose: bool = False):
        """
        Permanently deletes a record from the database.  Will issue an error 
        if exactly one matching record is not found in the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to delete from the database.  If not given, name and/or
            style are needed to uniquely define the record to delete.
        name : str, optional
            The name of the record to delete.
        style : str, optional
            The style of the record to delete.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
            
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """
        # Extract values from Record object if given
        if record is not None:
            if style is not None or name is not None:
                raise ValueError('kwargs style and name cannot be given with kwarg record')
            name = record.name
            style = record.style

        # Delete record
        self.cdcs.delete_record(template=style, title=name)

        if verbose:
            print(f'{record} deleted from {self.host}')

    def assign_records(self,
                       records: Union[Record, list],
                       workspace: Union[str, pd.Series],
                       verbose: bool = False):
        """
        Assigns one or more records to a CDCS workspace.

        Parameters
        ----------
        records : Record or list
            The record(s) to assign to the workspace.
        workspace : str or pandas.Series
            The workspace to assign the records to.
        verbose : bool, optional
            Setting this to True will print extra status messages.  Default
            value is False.
        """
        #ids = []
        for record in aslist(records):
            self.cdcs.assign_records(workspace, template=record.style,
                                     title=record.name)
            if verbose:
                print(f'{record} assigned to workspace {workspace}')

    def add_tar(self, 
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                tar: Optional[bytes] = None,
                root_dir: Optional[Path] = None):
        """
        Archives and stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has an archive.
        """

        # Get Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        else:
            # Issue a ValueError for competing kwargs
            if style is not None or name is not None:
                raise ValueError('kwargs style and name cannot be given with kwarg record')

            # Verify that record exists
            record = self.get_record(name=record.name, style=record.style)

        # Check if an archive already exists
        blobs = self.cdcs.get_blobs(filename=record.name)
        if len(blobs) > 0:
            raise ValueError('Record already has an archive')

        # Create directory archive and upload
        if tar is None: 
            if root_dir is None:
                root_dir = Path.cwd()

            # Make archive
            basename = Path(root_dir, record.name)
            filename = Path(root_dir, record.name + '.tar.gz')
            shutil.make_archive(basename, 'gztar', root_dir=root_dir,
                                base_dir=record.name)

            # Upload archive
            tries = 0
            while tries < 2:
                tries += 1
                try:
                    url = self.cdcs.upload_blob(filename.as_posix())
                    break
                except Exception as err:
                    if tries == 2:
                        raise ValueError('Failed to upload archive 2 times') from err

            # Remove local archive copy
            filename.unlink()

        # Upload pre-existing tar object
        elif root_dir is None:
            filename = Path(record.name + '.tar.gz')

            # Upload archive
            tries = 0
            while tries < 2:
                tries += 1
                try:
                    url = self.cdcs.upload_blob(filename=filename, blobbytes=BytesIO(tar))
                    break
                except Exception as err:
                    if tries == 2:
                        raise ValueError('Failed to upload archive 2 times') from err

        else:
            raise ValueError('tar and root_dir cannot both be given')

    def get_tar(self,
                record: Optional[Record] = None,
                style: Optional[str] = None,
                name: Optional[str] = None,
                raw: bool = False) -> Union[tarfile.TarFile, bytes]:
        """
        Retrieves the tar archive associated with a record in the database.
        
        Parameters
        ----------
        record : Record, optional
            The record to retrieve the associated tar archive for.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        raw : bool, optional
            If True, return the archive as raw binary content. If 
            False, return as an open tarfile. (Default is False)
            
        Returns
        -------
        tarfile or str
            The tar archive as an open tarfile if raw=False, or as a binary str if
            raw=True.
            
        Raises
        ------
        ValueError
            If style and/or name content given with record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a TypeError for competing kwargs
        elif style is not None or name is not None:
            raise TypeError('kwargs style and name cannot be given with kwarg record')

        filename = Path(record.name + '.tar.gz')

        # Download tar file
        tardata = self.cdcs.get_blob_contents(filename=filename)

        # Return contents
        if raw is True:
            return tardata
        else:
            tar = tarfile.open(fileobj = BytesIO(tardata))
            record.tar = tar
            return tar

    def delete_tar(self, 
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None):
        """
        Deletes a tar file from the database.
        
        Parameters
        ----------
        record : Record, optional
            The record associated with the tar archive to delete.  If not
            given, then name and/or style necessary to uniquely identify
            the record are needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        """

        # Create Record object if not given
        if record is None:
            record = self.get_record(name=name, style=style)

        # Issue a TypeError for competing kwargs
        elif style is not None or name is not None:
            raise TypeError('kwargs style and name cannot be given with kwarg record')

        filename = Path(record.name + '.tar.gz')

        self.cdcs.delete_blob(filename=filename)

    def update_tar(self,
                   record: Optional[Record] = None,
                   style: Optional[str] = None,
                   name: Optional[str] = None,
                   tar: Optional[bytes] = None,
                   root_dir: Optional[Path] = None):
        """
        Archives and stores a folder associated with a record.
        
        Parameters
        ----------
        record : Record, optional
            The record to associate the tar archive with.  If not given, then
            name and/or style necessary to uniquely identify the record are
            needed.
        name : str, optional
            The name to use in uniquely identifying the record.
        style : str, optional
            The style to use in uniquely identifying the record.
        tar : bytes, optional
            The bytes content of a tar file to save.  tar cannot be given
            with root_dir.
        root_dir : str, optional
            Specifies the root directory for finding the directory to archive.
            The directory to archive is at <root_dir>/<name>.  (Default is to
            set root_dir to the current working directory.)  tar cannot be given
            with root_dir.
        
        Raises
        ------
        ValueError
            If style and/or name content given with record or the record already
            has an archive.
        """

        self.delete_tar(record=record, name=name, style=style)
        self.add_tar(record=record, name=name, style=style, tar=tar, root_dir=root_dir)



@property
def cm(self) -> float:
    """centimeter"""
    return 1e-2 * self.m

@property
def mm(self) -> float:
    """millimeter"""
    return  1e-3 * self.m
    
@property
def um(self) -> float:
    """micrometer"""
    return 1e-6 * self.m
    
@property
def nm(self) -> float:
    """nanometer"""
    return 1e-9 * self.m
    
@property
def pm(self) -> float:
    """picometer"""
    return 1e-12 * self.m
    
@property
def fm(self) -> float:
    """femtometer"""
    return 1e-15 * self.m
    
@property
def km(self) -> float:
    """kilometer"""
    return 1e3 * self.m
    
@property
def angstrom(self) -> float:
    """angstrom"""
    return 1e-10 * self.m
    
@property
def Å(self) -> float:
    """angstrom"""
    return self.angstrom
    
@property
def lightyear(self) -> float:
    """lightyear"""
    return 9460730472580800. * self.m
    
@property
def astro_unit(self) -> float:
    """astronomical unit"""
    return 149597870700. * self.m 
    
@property
def pc(self) -> float:
    """parsec"""
    return (648000. / self.pi) * self.astro_unit
    
@property
def kpc(self) -> float:
    """kiloparsec"""
    return 1e3 * self.pc
    
@property
def Mpc(self) -> float:
    """megaparsec"""
    return 1e6 * self.pc
    
@property
def Gpc(self) -> float:
    """gigaparsec"""
    return 1e9 * self.pc
    
@property
def inch(self) -> float:
    """inch"""
    return 2.54 * self.cm
    
@property
def foot(self) -> float:
    """foot"""
    return 12. * self.inch
    
@property
def mile(self) -> float:
    """mile"""
    return 5280. * self.foot
    
@property
def thou(self) -> float:
    """thousandth of an inch, mil"""
    return 1e-3 * self.inch


# coding: utf-8
# Standard Python libraries
from pathlib import Path
from importlib import resources
from typing import Union, Optional
import io
from tarfile import TarFile

# https://ipython.org/
from IPython.display import display, HTML

# https://lxml.de/
import lxml.etree as ET

import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from .. import load_query

class Record():
    """
    Class for handling different record styles in the same fashion.  The
    base class defines the common methods and attributes.
    """

    def __init__(self,
                 model: Union[str, io.IOBase, DM, None] = None,
                 name: Optional[str] = None,
                 database = None,
                 **kwargs: any):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, or DataModelDict, optional
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        database : yabadaba.Database, optional
            A default Database to associate with the Record, typically the
            Database that the Record was obtained from.  Can allow for Record
            methods to perform Database operations without needing to specify
            which Database to use.
        kwargs : any
            Any record-specific attributes to assign.
        """
        self.__model = None
        self.__name = None
        self.tar = None
        self.database = database

        if model is not None:
            assert len(kwargs) == 0, f"cannot specify kwargs with model: '{kwargs.keys()}'"
            self.load_model(model, name=name)
        else:
            self.set_values(name=name, **kwargs)

    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str, file-like object, or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        # Get name if model is a filename
        if name is None:
            try:
                if Path(model).is_file():
                    self.name = Path(model).stem
            except (ValueError, OSError, TypeError):
                pass
        else:
            self.name = name

        self._set_model(model)

    def set_values(self, name: Optional[str] = None):
        """
        Set multiple object attributes at the same time.

        Parameters
        ----------
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        # Set name if given
        if name is not None:
            self.name = name

    def __str__(self) -> str:
        """str: The string representation of the record"""
        return f'{self.style} record named {self.name}'

    @property
    def style(self) -> str:
        """str: The record style"""
        return 'base'

    @property
    def xsd_filename(self) -> tuple:
        """tuple: The module path and file name of the record's xsd schema"""
        raise NotImplementedError('Not implemented')

    @property
    def xsd(self) -> bytes:
        """bytes: The xml schema for the record style."""
        return resources.read_binary(*self.xsd_filename)

    @property
    def xsl_filename(self) -> tuple:
        """tuple: The module path and file name of the record's xsl html transformer"""
        raise NotImplementedError('Not implemented')

    @property
    def xsl(self) -> bytes:
        """bytes: The xsl transformer for the record style."""
        return resources.read_binary(*self.xsl_filename)

    @property
    def name(self) -> str:
        """str: The record's name."""
        if self.__name is not None:
            return self.__name
        else:
            raise AttributeError('record name not set')

    @name.setter
    def name(self, value: Optional[str]):
        if value is not None:
            self.__name = str(value)
        else:
            self.__name = None

    @property
    def modelroot(self) -> str:
        """str : The name of the root element in the model contents."""
        raise NotImplementedError('Specific to subclasses')

    @property
    def model(self) -> DM:
        """DataModelDict: The record's model content."""
        if self.__model is not None:
            return self.__model
        else:
            raise AttributeError('model content has not been loaded or built')

    def reload_model(self):
        """
        Reloads the record based on the model content.  This allows for direct
        changes to the model to be updated to the object. 
        """
        self.load_model(model=self.model, name=self.name)

    def _set_model(self, model: DM):
        """
        Sets model content - called by build_model() and load_model() to update
        content.  Use load_model() if you are passing in an external model.
        """
        try:
            modelroot = self.modelroot
        except NotImplementedError:
            self.__model = DM(model)
        else:
            # Load model as DataModelDict
            content = DM(model).find(modelroot)
            self.__model = DM([(modelroot, content)])

    def build_model(self):
        """
        Generates and returns model content based on the values set to object.
        """
        raise NotImplementedError('Not defined for this class')

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        return {}

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        return {}

    @property
    def querynames(self) -> list:
        """list: The query parameter names supported by the record."""
        return list(self.queries.keys())

    @property
    def querydoc(self) -> str:
        """str: A description of all the queries supported by the record."""
        doc = f'# {self.style} Query Parameters\n\n'
        for name, query in self.queries.items():
            doc += f'- __{name}__ (*{query.parameter_type}*): {query.description}\n'

        return doc

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     **kwargs: any) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list, optional
            The record name(s) to parse by.
        **kwargs : any
            Any of the record style-specific search parameters.

        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        # Get defined queries
        queries = self.queries

        # Query name
        matches = load_query('str_match', name='name').pandas(dataframe, name)

        # Apply queries based on given kwargs
        for key in kwargs:
            matches = (matches & queries[key].pandas(dataframe, kwargs[key]))

        return matches

    def mongoquery(self,
                   name: Union[str, list, None] = None,
                   **kwargs: any) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.

        Parameters
        ----------
        name : str or list, optional
            The record name(s) to parse by.
        **kwargs : any
            Any of the record style-specific search parameters.

        Returns
        -------
        dict
            The Mongo-style query
        """
        # Get the dict of queries
        queries = self.queries

        # Initialize the full query dict and list of query operations
        querydict = {}
        querydict['$and'] = querylist = [{}]

        # Query name
        load_query('str_match', path='name').mongo(querylist, name)

        # Apply queries based on given kwargs
        for key in kwargs:
            queries[key].mongo(querylist, kwargs[key], prefix='content.')

        return querydict

    def cdcsquery(self,
                  **kwargs: any) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        **kwargs : any
            Any of the record style-specific search parameters.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        # Get the dict of queries
        queries = self.queries

        # Initialize the query dictionary
        querydict = {}
        querydict['$and'] = querylist = [{}]

        # Apply queries based on given kwargs
        for key in kwargs:
            queries[key].mongo(querylist, kwargs[key])

        return querydict

    def html(self,
             render: bool = False) -> Optional[str]:
        """
        Returns an HTML representation of the object.
        
        Parameters
        ----------
        render : bool, optional
            If True, then IPython is used to render the HTML.  If False
            (default), then the HTML code is returned as a str.

        Returns
        -------
        str
            The HTML code contents.  Returned if render=False.
        """

        # Build xml content
        xml_content = self.model.xml()

        xml = ET.fromstring(xml_content.encode('UTF-8'))

        # Read xsl content
        xsl = ET.fromstring(self.xsl)

        # Transform to html
        transform = ET.XSLT(xsl)
        html = transform(xml)
        html_content = ET.tostring(html).decode('UTF-8')

        if render:
            display(HTML(html_content))
        else:
            return html_content

    def valid_xml(self,
                  xml_content: Optional[str] = None) -> bool:
        """
        Tests if XML content is valid with schema.
        
        Parameters
        ----------
        xml_content : str, optional
            XML content to test against the record's schema.
            If not given, will generate the xml using build_model.
        
        Returns
        -------
        bool
            Indicating if XML is valid.
        """

        # Build xml content
        if xml_content is None:
            xml_content = self.model.xml()

        xml = ET.fromstring(xml_content.encode('UTF-8'))

        # Read xsd content
        xsd = ET.fromstring(self.xsd)

        schema = ET.XMLSchema(xsd)
        return schema.validate(xml)

    @property
    def database(self):
        """yabadaba.Database or None: The default Database associated with the Record"""
        return self.__database

    @database.setter
    def database(self, value):
        if value is None or hasattr(value, 'get_records'):
            self.__database = value
        else:
            raise TypeError('database must be a yabadaba.Database or None')

    @property
    def tar(self):
        """tarfile.TarFile: The tar archive associated with the record"""
        # Return tarfile if set
        if self.__tar is not None:
            return self.__tar

        # Check if database is set
        if self.database is None:
            raise ValueError('tar not loaded and no database set')

        # Fetch tar from database, set to cache and return
        self.tar = self.database.get_tar(record=self)
        return self.__tar

    @tar.setter
    def tar(self, value: Optional[TarFile]):
        if value is None or isinstance(value, TarFile):
            self.__tar = value
        else:
            raise TypeError('tar must ne a TarFile or None')

    def clear_tar(self):
        """Closes and unsets the record's tar file to save memory"""
        if self.__tar is not None:
            self.__tar.close()
            self.tar = None

    def get_file(self,
                 filename: Union[str, Path],
                 localroot: Union[str, Path, None] = None):
        """
        Retrieves a file either locally or from the record's tar archive.

        Parameters
        ----------
        filename : str or Path
            The name/path for the file.  For local files, this is taken
            relative to localroot.  For files in the tar archive, this is taken
            relative to the tar's root directory which is always named for the
            record, i.e., {self.name}/{filename}.
        localroot : str, Path or None, optional
            The local root directory that filename (if it exists) is relative
            to.  The default value of None will use the current working
            directory.
        
        Raises
        ------
        ValueError
            If filename exists in the tar but is not a file.

        Returns
        -------
        io.IOBase
            A file-like object in binary read mode that allows for the file
            contents to be read.
        """
        # Set default root path
        if localroot is None:
            localroot = Path.cwd()
        else:
            localroot = Path(localroot)

        # Return local copy of file if it exists
        localfile = Path(localroot, filename)
        if Path(localfile).is_file():
            return open(localfile, 'rb')

        # Return file extracted from tar
        fileio = self.tar.extractfile(f'{self.name}/{filename}')
        if fileio is not None:
            return fileio
        else:
            raise ValueError(f'{filename} exists in tar, but is not a file')



from yabadaba.record import Record
from yabadaba import load_query

from DataModelDict import DataModelDict as DM

class FAQ(Record):
    """
    Class for representing FAQ (frequently asked question) records.
    """
    def __init__(self, model=None, name=None, **kwargs):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, DataModelDict
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        """
        self.__question = None
        self.__answer = None
        super().__init__(model=model, name=name, **kwargs)

    @property
    def style(self):
        """str: The record style"""
        return 'FAQ'

    @property
    def modelroot(self):
        """str: The root element of the content"""
        return 'faq'
    
    @property
    def xsl_filename(self):
        """tuple: The module path and file name of the record's xsl html transformer"""
        return ('yabadaba.demo', 'FAQ.xsl')

    @property
    def xsd_filename(self):
        """tuple: The module path and file name of the record's xsd schema"""
        return ('yabadaba.demo', 'FAQ.xsd')

    @property
    def question(self):
        """str: The frequently asked question."""
        return self.__question

    @question.setter
    def question(self, value):
        if value is None:
            self.__question = None
        else:
            self.__question = str(value)

    @property
    def answer(self):
        """str: The answer to the frequently asked question."""
        return self.__answer

    @answer.setter
    def answer(self, value):
        if value is None:
            self.__answer = None
        else:
            self.__answer = str(value)

    def load_model(self, model, name=None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)

        faq = self.model[self.modelroot]
        self.question = faq['question']
        self.answer = faq['answer']

    def set_values(self, name=None, question=None, answer=None):
        """
        Set multiple object attributes at the same time.

        Parameters
        ----------
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        question : str, optional
            The frequently asked question.
        answer : str, optional
            The answer to the frequently asked question.
        """
        if question is not None:
            self.question = question
        if answer is not None:
            self.answer = answer
        if name is not None:
            self.name = name

    def build_model(self):
        """
        Generates and returns model content based on the values set to object.
        """
        model = DM()
        model['faq'] = DM()
        model['faq']['question'] = self.question
        model['faq']['answer'] = self.answer

        self._set_model(model)
        return model

    def metadata(self):
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        meta = {}
        meta['name'] = self.name
        meta['question'] = self.question
        meta['answer'] = self.answer
        return meta

    @property
    def queries(self):
        """dict: Query objects and their associated parameter names."""
        return {
            'question': load_query('str_contains',
                name='question', path=f'{self.modelroot}.question',
                description='Search the FAQ question field to see if it contains certain strings.'),
            'answer': load_query('str_contains',
                name='answer', path=f'{self.modelroot}.answer',
                description='Search the FAQ answer field to see if it contains certain strings.')
        }



@property
def mol(self) -> float:
    """mole"""
    return self.NA

@property
def mmol(self) -> float:
    """millimole"""
    return 1e-3 * self.mol

@property
def umol(self) -> float:
    """micromole"""
    return 1e-6 * self.mol

@property
def nmol(self) -> float:
    """nanomole"""
    return 1e-9 * self.mol

@property
def pmol(self) -> float:
    """picomole"""
    return 1e-12 * self.mol

@property
def fmol(self) -> float:
    """femtomole"""
    return 1e-15 * self.mol

@property
def M(self) -> float:
    """molar"""
    return self.mol / self.L

@property
def mM(self) -> float:
    """millimolar"""
    return 1e-3 * self.M

@property
def uM(self) -> float:
    """micromolar"""
    return 1e-6 * self.M

@property
def nM(self) -> float:
    """nanomolar"""
    return 1e-9 * self.M

@property
def pM(self) -> float:
    """picomolar"""
    return 1e-12 * self.M

@property
def fM(self) -> float:
    """femtomolar"""
    return 1e-15 * self.M


# coding: utf-8
# Standard Python libraries
from typing import Optional

# Relative imports
from ..tools import screen_input, ModuleManager
from . import databasemanager as default_databasemanager
from . import Database
from .. import settings as default_settings
from .. import Settings

__all__ = ['load_database']

def load_database(name: Optional[str] = None,
                  style: Optional[str] = None,
                  host: Optional[str] = None,
                  settings: Optional[Settings.Settings] = None, 
                  databasemanager: Optional[ModuleManager] = None,
                  **kwargs) -> Database:
    """
    Loads a database object.  Can be either loaded from stored settings or
    by defining all needed access information.
    
    Parameters
    ----------
    name : str, optional
        The name assigned to a pre-defined database.  If given, can be the only
        parameter.
    style : str, optional
        The database style to use.
    host : str, optional
        The URL/file path where the database is hosted.
    settings : yabadaba.Settings.Settings, optional
        A Settings object.  Allows for different settings files to be used
        by downstream packages.
    databasemanager : yabadaba.tools.ModuleManager, optional
        Allows for an alternate databasemanager to be specified by downstream
        packages. 
    kwargs : dict, optional
        Any other keyword parameters defining necessary access information.
        Allowed keywords are database style-specific.
    
    Returns
    -------
    Subclass of datamodelbase.Database
        The database object.
    """

    # Check that style and name are both not given
    if style is not None and name is not None:
        raise ValueError('name and style cannot both be given')

    # Set default package values 
    if settings is None:
        settings = default_settings
    if databasemanager is None:
        databasemanager = default_databasemanager

    # Load Database info from saved settings
    if style is None:

        if host is not None or len(kwargs) > 0:
            raise ValueError('style is required if host and/or kwargs are given')

        # Get information from settings file
        database_names = settings.list_databases
        
        # Ask for name if not given
        if name is None:
            if len(database_names) > 0:
                print('Select a database:')
                for i, database in enumerate(database_names):
                    print(i+1, database)
                choice = screen_input(':')
                try:
                    choice = int(choice)
                except ValueError:
                    name = choice
                else:
                    name = database_names[choice-1]
            else:
                raise KeyError('No databases currently set')
        
        try:
            kwargs = settings.databases[name]
        except KeyError as err:
            raise KeyError(f'database {name} not found') from err

        style = kwargs.pop('style')
        host = kwargs.pop('host')

    return databasemanager.init(style, host=host, **kwargs)



# coding: utf-8

# Standard Python libraries
from typing import Any, Optional

import pandas as pd

# Relative imports
from ..tools import aslist, iaslist
from .Query import Query

class StrMatchQuery(Query):
    """Class for querying str fields for specific values"""

    @property
    def style(self) -> str:
        """str: The query style"""
        return 'str_match'

    @property
    def parameter_type(self) -> str:
        """str: The types of query parameter values accepted by this query style"""
        return 'str or list, optional'

    def mongo(self,
              querylist: list,
              value: Any,
              prefix: str = ''):
        """
        Builds a Mongo query operation for the field.

        Parameters
        ----------
        querylist : list
            The working list of mongo query operations which is to be appended
            with the operation for this query object.
        value : any
            The value of the field to query on.  If None, then no new query
            operation will be added.
        prefix : str, optional
            An optional prefix to add before the query path.  Used by Record's
            mongoquery to start each path with "content."
        """
        # Get path and add prefix
        path = f'{prefix}{self.path}'

        if value is not None:

            # Build the query 
            querylist.append( {path: {'$in': aslist(value)} } )

    def pandas(self,
               df: pd.DataFrame,
               value: Any) -> pd.Series:
        """
        Applies a query filter to the metadata for the field.
        
        Parameters
        ----------
        df : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        value : any
            The value of the field to query on.  If None, then it should return
            True for all rows of df.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """

        def apply_function(series: pd.Series,
                           name: str,
                           value: Any,
                           parent: Optional[str]) -> bool:
            """
            function for pandas.DataFrame.apply with axis=1
            
            Parameters
            ----------
            series : pd.Series
                A series of the DataFrame being operated on.
            name : str
                The element name.
            value : any
                The values to search for.
            parent : str or None
                The parent element name, if there is one.

            Returns
            -------
            bool
                True if value is None or if one given value matches the
                element being checked.
            """
            # Return True for all fields if value is None
            if value is None:
                return True
            
            if parent is None:
                
                # Check if name is in series
                if name not in series or pd.isna(series[name]):
                    return False

                # Check for a value match
                return series[name] in aslist(value)
            
            else:

                 # Loop over all child elements
                for child in iaslist(series[parent]):

                    # Check if child element has name
                    if name in child and pd.notna(child[name]):
                        
                        # Check if child element matches a value
                        if child[name] in aslist(value):
                            return True
                
                # Return default False for no matching child elements
                return False

        # Use apply_function on df using value and object attributes
        return df.apply(apply_function, axis=1, args=(self.name, value, self.parent))



from setuptools import setup, find_packages

def getreadme():
    """Fetches long description from the README file"""
    with open('README.rst', encoding='utf-8') as readme_file:
        return readme_file.read()

def getversion():
    """Fetches version information from VERSION file"""
    with open('yabadaba/VERSION', encoding='utf-8') as version_file:
        return version_file.read().strip()

setup(name = 'yabadaba',
      version = getversion(),
      description = 'Yay, a base database! An abstraction layer allowing for common interactions with Mongo, CDCS and local directory databases and records.',
      long_description = getreadme(),
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Physics'
      ],
      keywords = [
        'database', 
        'mongodb', 
        'CDCS',
      ],
      url = 'https://github.com/usnistgov/yabadaba',
      author = 'Lucas Hale',
      author_email = 'lucas.hale@nist.gov',
      packages = find_packages(),
      install_requires = [
        'lxml',
        'DataModelDict',
        'IPython',
        'numpy', 
        'pandas',
        'cdcs>=0.2.3',
        'pymongo',
        'tqdm',
      ],
      package_data={'': ['*']},
      zip_safe = False)



@property 
def V(self) -> float:
    """volt"""
    return self.J / self.C

@property
def mV(self) -> float:
    """millivolt"""
    return 1e-3 * self.V

@property
def uV(self) -> float:
    """microvolt"""
    return 1e-6 * self.V

@property
def nV(self) -> float:
    """nanovolt"""
    return 1e-9 * self.V

@property
def kV(self) -> float:
    """kilovolt"""
    return 1e3 * self.V

@property
def MV(self) -> float:
    """megavolt"""
    return 1e6 * self.V

@property
def GV(self) -> float:
    """gigavolt"""
    return 1e9 * self.V

@property
def TV(self) -> float:
    """teravolt"""
    return 1e12 * self.V


@property
def Pa(self) -> float:
    """pascal"""
    return self.N / self.m**2

@property
def hPa(self) -> float:
    """hectopascal"""
    return 1e2 * self.Pa

@property
def kPa(self) -> float:
    """kilopascal"""
    return 1e3 * self.Pa

@property
def MPa(self) -> float:
    """megapascal"""
    return 1e6 * self.Pa

@property
def GPa(self) -> float:
    """gigapascal"""
    return 1e9 * self.Pa

@property
def bar(self) -> float:
    """bar"""
    return 1e5 * self.Pa

@property
def mbar(self) -> float:
    """millibar"""
    return 1e-3 * self.bar

@property
def cbar(self) -> float:
    """centibar"""
    return 1e-2 * self.bar

@property
def dbar(self) -> float:
    """decibar"""
    return 0.1 * self.bar

@property
def kbar(self) -> float:
    """kilobar"""
    return 1e3 * self.bar

@property
def Mbar(self) -> float:
    """megabar"""
    return 1e6 * self.bar

@property
def atm(self) -> float:
    """standard atmosphere pressure"""
    return 101325. * self.Pa

@property
def torr(self) -> float:
    """torr"""
    return (1. / 760.) * self.atm

@property
def mtorr(self) -> float:
    """millitorr"""
    return 1e-3 * self.torr

@property
def psi(self) -> float:
    """pounds force per square inch"""
    return self.lbf / self.inch**2


from yabadaba.record import Record

from DataModelDict import DataModelDict as DM

raise NotImplementedError('BadRecord is meant to fail import!')
class BadRecord(Record):
    """
    Class for representing FAQ (frequently asked question) records.
    """
    @property
    def style(self):
        """str: The record style"""
        return 'BadRecord'

    @property
    def modelroot(self):
        """str: The root element of the content"""
        return 'bad-record'

    def load_model(self, model, name=None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)

        bad = self.model[self.modelroot]

    def set_values(self, name=None):
        """
        Set multiple object attributes at the same time.

        Parameters
        ----------
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        if name is not None:
            self.name = name

    def build_model(self):
        """
        Generates and returns model content based on the values set to object.
        """
        model = DM()
        model['bad'] = DM()

        self._set_model(model)
        return model

    def metadata(self):
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        meta = {}
        meta['name'] = self.name
        return meta


# coding: utf-8

# Standard Python libraries
from typing import Any, Optional

import numpy as np

import pandas as pd

# Relative imports
from .. import unitconvert as uc
from ..tools import iaslist

from .Query import Query

class FloatMatchQuery(Query):
    """Class for querying float fields for closely matching values"""

    def __init__(self,
                 name: Optional[str] = None,
                 parent: Optional[str] = None,
                 path: Optional[str] = None,
                 description: str = '',
                 unit: Optional[str] = None,
                 atol: float = 1e-5):
        """
        Query initialization

        Parameters
        ----------
        name : str or None, optional
            The metadata key associated with the data field.  Must be set
            to use the pandas query method.
        parent : str or None, optional
            Allows for the pandas query operations to work on embedded
            metadata dicts.  If given, the pandas query method will check the
            value of metadata[parent][name].
        path : str or None, optional
            The record data path to the data field.  Levels are delimited by
            periods.  Must be given to use the mongo query method.
        description : str, optional
            Description of the query operation, i.e. what it is searching.
        unit : str or None, optional
            Unit to use for the given query value.  This should be the unit
            used in the JSON/XML records for the value in question.  A value
            of None indicates no unit or conversion needed.
        atol : float, optional
            Absolute tolerance for comparison.  Returned True values will be
            for float fields between +-atol of the target value.  Default
            value is 1e-5.
        """
        self.__atol = atol
        self.__unit = unit
        super().__init__(name=name, parent=parent, path=path, description=description)

    @property
    def style(self) -> str:
        """str: The query style"""
        return 'float_match'

    @property
    def parameter_type(self) -> str:
        """str: The types of query parameter values accepted by this query style"""
        return 'float or list, optional'

    @property
    def unit(self) -> str:
        """str : The unit to use for the query value"""
        return self.__unit

    @property
    def atol(self) -> float:
        """float: The absolute tolerance to use for value comparison"""
        return self.__atol

    def mongo(self,
              querylist: list,
              value: Any,
              prefix: str = ''):
        """
        Builds a Mongo query operation for the field.

        Parameters
        ----------
        querylist : list
            The working list of mongo query operations which is to be appended
            with the operation for this query object.
        value : any
            The value of the field to query on.  If None, then no new query
            operation will be added.
        prefix : str, optional
            An optional prefix to add before the query path.  Used by Record's
            mongoquery to start each path with "content."
        """
        # Get path and add prefix
        path = f'{prefix}{self.path}'

        if value is not None:

            # Init newquery as list of $or evaluations
            newquery = {'$or':[]}

            # Iterate over all values
            for v in iaslist(value):
                v = float(v)

                # Add query operation for ranged search
                newquery['$or'].append({path:{'$gte': v - self.atol, '$lte' : v + self.atol}})
            
            # Append newquery to querylist
            querylist.append(newquery)

    def pandas(self,
               df: pd.DataFrame,
               value: Any) -> pd.Series:
        """
        Applies a query filter to the metadata for the field.
        
        Parameters
        ----------
        df : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        value : any
            The value of the field to query on.  If None, then it should return
            True for all rows of df.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """

        def apply_function(series: pd.Series,
                           name: str,
                           value: Any,
                           parent: Optional[str]) -> bool:
            """
            function for pandas.DataFrame.apply with axis=1
            
            Parameters
            ----------
            series : pd.Series
                A series of the DataFrame being operated on.
            name : str
                The element name.
            value : any
                The values to search for.
            parent : str or None
                The parent element name, if there is one.

            Returns
            -------
            bool
                True if value is None or if one given value matches the
                element being checked.
            """
            # Return True for all fields if value is None
            if value is None:
                return True
            
            # Convert value to array of floats
            value = np.array([float(v) for v in iaslist(value)])

            # Convert value to target units
            value = uc.set_in_units(value, self.unit)

            if parent is None:

                # Check if name is in series
                if name not in series or pd.isna(series[name]):
                    return False
                
                # Check for a value match
                return np.any(np.isclose(float(series[name]), value, rtol=0.0, atol=self.atol))
            
            else:

                # Loop over all child elements
                for child in iaslist(series[parent]):

                    # Check if child element has name
                    if name in child and pd.notna(child[name]):

                        # Check if child element matches a value
                        if np.any(np.isclose(float(child[name]), value, rtol=0.0, atol=self.atol)):
                            return True

                # Return default False for no matching child elements
                return False

        # Use apply_function on df using value and object attributes
        return df.apply(apply_function, axis=1, args=(self.name, value, self.parent))
